@@ -11,14 +11,14 @@
 
    using python, ast module to parse a python expression.
 
-   version: 0.9.1
-   last update: 9-July-2012
+   version: 0.9.2
+   last update: 13-Feb-2014
    License:  BSD
    Author:  Matthew Newville <newville@cars.uchicago.edu>
             Center for Advanced Radiation Sources,
             The University of Chicago
 """
-__version__ = '0.9.1'
+__version__ = '0.9.2'
 
 from .asteval import Interpreter
 from .astutils import NameFinder, valid_symbol_name
